(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Context_Substitution
imports
  Trs
  "~~/src/HOL/Library/Monad_Syntax"
  Substitution
begin


fun ctxt_subst :: "('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> nat \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
  where "ctxt_subst _ _ 0 t = t"
     |  "ctxt_subst C \<mu> (Suc n) t = C\<langle>(ctxt_subst C \<mu> n t) \<cdot> \<mu>\<rangle>"

fun ctxt_subst_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt"
  where "ctxt_subst_ctxt _ _ 0 D = D"
     |  "ctxt_subst_ctxt C \<mu> (Suc n) D = C \<circ>c (ctxt_subst_ctxt C \<mu> n D \<cdot>c \<mu>)"


lemma ctxt_subst_ctxt: "ctxt_subst C \<mu> n (D\<langle>t\<rangle>) = (ctxt_subst_ctxt C \<mu> n D)\<langle>t \<cdot> \<mu>^^n\<rangle>"
  by (induct n, auto simp: subst_left_right)

lemma ctxt_subst_add[simp]: "ctxt_subst C \<mu> (n + m) t = ctxt_subst C \<mu> n (ctxt_subst C \<mu> m t)"
  by (induct n, auto)

lemma ctxt_subst_ctxt_add[simp]: "ctxt_subst_ctxt C \<mu> (n + m) D = ctxt_subst_ctxt C \<mu> n (ctxt_subst_ctxt C \<mu> m D)"
  by (induct n, auto)

lemma ctxt_subst_Suc: "ctxt_subst C \<mu> (Suc n) t = ctxt_subst C \<mu> n (C\<langle>t\<cdot>\<mu>\<rangle>)" (is "?l = _")
proof -
  have "?l = ctxt_subst C \<mu> (n + Suc 0) t" by simp
  from this[unfolded ctxt_subst_add] show ?thesis by simp
qed

lemma ctxt_subst_ctxt_Suc: "ctxt_subst_ctxt C \<mu> (Suc n) D = ctxt_subst_ctxt C \<mu> n (C \<circ>c (D \<cdot>c \<mu>))" (is "?l = _")
proof -
  have "?l = ctxt_subst_ctxt C \<mu> (n + Suc 0) D" by simp
  from this[unfolded ctxt_subst_ctxt_add] show ?thesis by simp
qed

lemma ctxt_subst_hole_pos[simp]: "hole_pos C ^ n \<in> poss (ctxt_subst C \<mu> n t)"
  by (induct n, auto)

lemma ctxt_subst_ctxt_hole_pos[simp]: "hole_pos (ctxt_subst_ctxt C \<mu> n D) = hole_pos C ^ n <#> hole_pos D"
  by (induct n, auto)

lemma ctxt_subst_subst: "(ctxt_subst C \<mu> n t) \<cdot> \<mu> = ctxt_subst (C \<cdot>c \<mu>) \<mu> n (t \<cdot> \<mu>)"
  by (induct n, auto) 

lemma ctxt_subst_subst_pow: "(ctxt_subst C \<mu> n t) \<cdot> \<mu> ^^ k= ctxt_subst (C \<cdot>c (\<mu> ^^ k)) \<mu> n (t \<cdot> (\<mu> ^^ k))"
 proof (induct k, simp)
  case (Suc j)
   have "(ctxt_subst C \<mu> n t) \<cdot> \<mu> ^^ (Suc j) = ((ctxt_subst C \<mu> n t) \<cdot> \<mu> ^^ j) \<cdot> \<mu>" using subst_right_left by auto
   also have "\<dots> =  ctxt_subst (C \<cdot>c (\<mu> ^^ j)) \<mu> n (t \<cdot> (\<mu> ^^ j)) \<cdot> \<mu> " using Suc by auto
   also have "\<dots> =  ctxt_subst (C \<cdot>c (\<mu> ^^ j) \<cdot>c \<mu>) \<mu> n (t \<cdot> (\<mu> ^^ j) \<cdot> \<mu>)" using ctxt_subst_subst by auto
   finally show ?case using ctxt_compose_subst_compose_distrib subst_compose_apply_term_distrib subst_power_Suc by metis
 qed

lemma ctxt_subst_subt_at[simp]: "(ctxt_subst C \<mu> n t) |_ (hole_pos C ^ n) = t \<cdot> \<mu>^^n"
proof (induct n)
  case (Suc n)
  let ?p = "hole_pos C"
  have "(ctxt_subst C \<mu> (Suc n) t) |_ (?p ^ Suc n) = C\<langle>ctxt_subst C \<mu> n t \<cdot> \<mu>\<rangle> |_ ?p <#> ?p ^ n" by simp
  also have "... = (ctxt_subst C \<mu> n t \<cdot> \<mu>) |_ (?p ^ n)" using subt_at_append[OF hole_pos_poss[of C]] by simp
  also have "... = ((ctxt_subst C \<mu> n t) |_ (?p ^ n)) \<cdot> \<mu>" by (rule subt_at_subst[OF ctxt_subst_hole_pos])
  also have "... = (t \<cdot> \<mu>^^n) \<cdot> \<mu>" unfolding Suc by simp
  finally show ?case using subst_left_right by simp
qed simp

lemma ctxt_subst_ctxt_of_pos_term: assumes q: "q \<in> poss s"
  shows "ctxt_subst_ctxt C \<mu> n (ctxt_of_pos_term q s) = ctxt_of_pos_term ((hole_pos C)^n <#> q) (ctxt_subst C \<mu> n s)"
proof (induct n)
  case (Suc n)
  have hp: "hole_pos C ^ Suc n = hole_pos C <#> hole_pos C ^ n" by auto
  have poss: "hole_pos C ^ n <#> q \<in> poss (ctxt_subst C \<mu> n s)"
    unfolding poss_append_poss using q by auto
  show ?case 
    unfolding ctxt_subst_ctxt.simps
    unfolding Suc
    unfolding ctxt_subst.simps
    unfolding hp 
    by (simp add: context_of_pos_term_append[OF hole_pos_poss] ctxt_of_pos_term_subst[OF poss])
qed simp

lemma ctxt_subst_step: fixes C :: "('f,'v)ctxt"
  defines p: "p \<equiv> hole_pos C"
  assumes step: "(s,t) \<in> rstep_r_p_s R lr q \<sigma>"
  shows "(ctxt_subst C \<mu> n s, ctxt_subst C \<mu> n t) \<in> rstep_r_p_s R lr ((p^n) <#> q) (\<sigma> \<circ>s \<mu>^^n)"
proof -
  obtain l r where l_r: "lr = (l,r)" by force
  from step[unfolded rstep_r_p_s_def' l_r]
  have q: "q \<in> poss s" and lr: "(l,r) \<in> R" and s: "s |_ q = l \<cdot> \<sigma>" and t: "t = replace_at s q (r \<cdot> \<sigma>)" by auto
  show ?thesis unfolding rstep_r_p_s_def'
  proof(rule, unfold split l_r fst_conv snd_conv, intro conjI)
    show "(l,r) \<in> R" by (rule lr)
  next
    show "p ^ n <#> q \<in> poss (ctxt_subst C \<mu> n s)" 
      unfolding poss_append_poss p
      using q by auto
  next
    show "(ctxt_subst C \<mu> n s) |_ p ^ n <#> q = l \<cdot> \<sigma> \<circ>s \<mu> ^^ n"
      unfolding p
      unfolding subt_at_append[OF ctxt_subst_hole_pos]
      unfolding ctxt_subst_subt_at
      unfolding subt_at_subst[OF q] s
      by simp
  next
    show "ctxt_subst C \<mu> n t = (replace_at (ctxt_subst C \<mu> n s) (p ^ n <#> q) (r \<cdot> \<sigma> \<circ>s \<mu> ^^ n))" (is "?l = ?r")
    proof -        
      have "?l = ctxt_subst C \<mu> n (replace_at s q (r \<cdot> \<sigma>))" unfolding t ..
      also have "... = ?r" unfolding ctxt_subst_ctxt p unfolding ctxt_subst_ctxt_of_pos_term[OF q] by simp
      finally show "?l = ?r" .
    qed
  qed
qed

end
