(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multihole_Context
imports 
  Trs 
  Util
  Term_Impl
  "~~/src/HOL/Library/Monad_Syntax"
begin

section {* multihole-contexts *}

datatype ('f,'v)mctxt =  MVar 'v | MHole | MFun 'f "('f,'v)mctxt list"

lemma
  fixes P :: "('f, 'v)mctxt \<Rightarrow> bool"
  assumes "\<And>x. P(MVar x)" and "P MHole" and "\<And>f cs. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P(MFun f cs)"
  shows mctxt_induct[case_names MVar MHole MFun, induct type: "mctxt"]:
    "P c" and "\<And>c. c \<in> set cs \<Longrightarrow> P c"
using assms by (induct c and cs) auto

fun num_holes :: "('f,'v)mctxt \<Rightarrow> nat" where
  "num_holes (MVar _) = 0"
| "num_holes MHole = 1"
| "num_holes (MFun _ ctxts) = listsum (map num_holes ctxts)"

fun fill_holes :: "('f,'v)mctxt \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
  "fill_holes (MVar x) _ = Var x"
| "fill_holes MHole [t] = t"
| "fill_holes (MFun f cs) ts = (Fun f (map (\<lambda> i. fill_holes (cs ! i) 
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

lemma fill_holes_inj: assumes "fill_holes C ss = fill_holes C ts"
  and ss: "num_holes C = length ss"
  and ts: "length ts = length ss"
  shows "ss = ts"
  using assms
proof (induct C arbitrary: ss ts)
  case (MHole ss ts)
  thus ?case by (cases ss, (cases ts, auto)+)
next
  case (MFun ff Cs ss ts)
  let ?s = "\<lambda> C ss. fill_holes C (take (num_holes C) ss)"
  let ?t = "\<lambda> i Cs ss. fill_holes (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ss))"
  let ?p = "\<lambda> i Cs ss ts. ?t i Cs ss = ?t i Cs ts"
  from MFun(2) have p: "\<And> i. i < length Cs \<Longrightarrow> ?p i Cs ss ts" by auto
  from MFun(3) have nh: "listsum (map num_holes Cs) = length ss" by auto
  from MFun(4) have l: "length ts = length ss" by auto
  from MFun(1) p nh l show ?case
  proof (induct Cs arbitrary: ss ts)
    case (Cons C Cs ss ts)
    note oIH = Cons(2)[of C]
    let ?n = "num_holes C"
    from Cons(3)[of 0] have id: "?s C ss = ?s C ts" by auto      
    from Cons(4) Cons(5) have n: "?n = length (take ?n ss)" "length (take ?n ts) = length (take ?n ss)" by auto
    from oIH[OF _ id n] have id: "take ?n ss = take ?n ts" by auto
    have idss: "\<And> f. f ss = f (take ?n ss @ drop ?n ss)" by simp
    have idts: "\<And> f. f ts = f (take ?n ss @ drop ?n ts)" using append_take_drop_id[of ?n ts, unfolded id[symmetric]] by auto
    have "(ss = ts) \<longleftrightarrow> take ?n ss @ drop ?n ss = take ?n ts @ drop ?n ts" by auto
    also have "\<dots> \<longleftrightarrow> drop ?n ss = drop ?n ts" unfolding id by (simp del: append_take_drop_id)
    also have "\<dots>"
    proof (rule Cons(1))
      fix C ss ts
      assume C: "C \<in> set Cs"
      and *: "fill_holes C ss = fill_holes C ts" "num_holes C = length ss" "length ts = length ss"
      from Cons(2)[OF _ *] C show "ss = ts" by auto
    next
      show "listsum (map num_holes Cs) = length (drop (num_holes C) ss)"
        using Cons(4) n(1) by auto
      show "length (drop (num_holes C) ts) = length (drop (num_holes C) ss)"
        using Cons(5)[unfolded idts[of length] idss[of length]] by auto
    next
      fix i
      assume "i < length Cs"
      hence i: "Suc i < length (C # Cs)" by simp
      show "?p i Cs (drop ?n ss) (drop ?n ts)"
        using Cons(3)[OF i] by (auto simp: ac_simps)
    qed
    finally show ?case by simp
  qed auto
qed auto

lemma fill_holes_arbitrary: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> f (Cs ! i) (ss ! i) = ts ! i"
  shows "map (\<lambda>i. f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))) [0..<length Cs] 
    = ts"
proof -
  have l: "length [0 ..< length Cs] = length ts" unfolding lCs by simp
  show ?thesis 
  proof (unfold map_nth_eq_conv[OF l], intro allI impI)
    fix i
    assume i: "i < length ts"
    with lCs have iC: "i < length Cs" by simp
    hence "f (Cs ! ([0..<length Cs] ! i))
         (take (num_holes (Cs ! ([0..<length Cs] ! i))) (drop (listsum (map num_holes (take ([0..<length Cs] ! i) Cs))) (concat ss)))
       = f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))" (is "?l = _")
     by simp
    also have "\<dots> = ts ! i" using i lCs lss rec
    proof (induct i arbitrary: ts Cs ss)
      case 0
      from 0 obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from 0 obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from 0 obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      show ?case using 0(4)[of 0] unfolding ts ss Cs by auto
    next
      case (Suc i)
      from Suc obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from Suc obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from Suc obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      from Suc ts ss Cs have i: "i < length ts'" "length Cs' = length ts'" "length ss' = length ts'" by auto
      from Suc(5)[of 0] ts ss Cs have nh: "num_holes C = length s" "f C s = t" by auto
      show ?case using Suc(5) unfolding Cs ss ts using nh
        by (force intro!: Suc(1) i)
    qed
    finally show "?l = ts ! i" .
  qed
qed

lemma fill_holes_MFun: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes (MFun f Cs) (concat ss) = Fun f ts" 
  unfolding fill_holes.simps term.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

inductive eq_fill :: "('f,'v)term \<Rightarrow> ('f,'v)mctxt \<times> ('f,'v)term list \<Rightarrow> bool" ("(_/ =f _)" [51,51] 50) where
  eqfI[intro]: "t = fill_holes D ss \<Longrightarrow> num_holes D = length ss \<Longrightarrow> t =f (D,ss)"

lemma eqfE: assumes "t =f (D,ss)" shows "t = fill_holes D ss" "num_holes D = length ss"
  using assms[unfolded eq_fill.simps] by auto

lemma eqf_MFunI: assumes "length sss = length Cs"
  "length ts = length Cs"
  "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
  shows "Fun f ts =f (MFun f Cs,concat sss)"
proof 
  have "num_holes (MFun f Cs) = listsum (map num_holes Cs)" by simp
  also have "map num_holes Cs = map length sss"
    by (rule nth_equalityI, insert assms eqfE[OF assms(3)], auto)
  also have "listsum (\<dots>) = length (concat sss)" unfolding length_concat ..
  finally show "num_holes (MFun f Cs) = length (concat sss)" .
  show "Fun f ts = fill_holes (MFun f Cs) (concat sss)"
    by (rule fill_holes_MFun[symmetric], insert assms(1,2) eqfE[OF assms(3)], auto)
qed

lemma eqf_MFunE: assumes "s =f (MFun f Cs,ss)"  
  obtains ts sss where "s = Fun f ts" "length ts = length Cs" "length sss = length Cs" 
  "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
  "ss = concat sss" 
proof -
  from eqfE[OF assms] have fh: "s = fill_holes (MFun f Cs) ss" 
    and nh: "listsum (map num_holes Cs) = length ss" by auto 
  from fh obtain ts where s: "s = Fun f ts" by (cases s, auto)
  from fh[unfolded s] 
  have ts: "ts = map (\<lambda>i. fill_holes (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ss))) [0..<length Cs]" 
    (is "_ = map (?f Cs ss) _")
    by auto
  from nh 
  have "\<exists> sss. length sss = length Cs \<and> (\<forall> i < length Cs . ts ! i =f (Cs ! i, sss ! i)) \<and> ss = concat sss"
    unfolding ts
  proof (induct Cs arbitrary: ss)
    case (Cons C Cs ss)
    let ?n = "num_holes C"
    let ?bef = "take ?n ss"
    let ?aft = "drop ?n ss"
    def bef \<equiv> ?bef
    def aft \<equiv> ?aft
    from Cons(2) have "length ?bef = ?n" "listsum (map num_holes Cs) = length ?aft" by auto
    hence lb: "?n = length bef" "listsum (map num_holes Cs) = length aft" unfolding bef_def aft_def by auto
    have ss: "ss = bef @ aft" unfolding bef_def aft_def by auto
    have l: "length (C # Cs) = Suc (length Cs)" by simp
    note IH = Cons(1)[OF lb(2)]
    from IH obtain sss where len: "length sss = length Cs" and
      i: "\<And> i. i < length Cs \<Longrightarrow> map (?f Cs aft) [0 ..< length Cs] ! i =f (Cs ! i, sss ! i)" and
      aft: "aft = concat sss" by blast
    have "?case \<longleftrightarrow> (\<exists>sss. length sss = Suc (length Cs) \<and>
          fill_holes C bef =f (C, sss ! 0) \<and>
          (\<forall>i<length Cs.
              ?f Cs aft i =f (Cs ! i, sss ! Suc i)) \<and>
          bef @ aft = concat sss)"
      unfolding l map_upt_Suc all_Suc_conv 
      unfolding ss by (simp add: lb)
    also have \<dots>
    proof (rule exI[of _ "bef # sss"], intro conjI allI impI)
      show "length (bef # sss) = Suc (length Cs)" by (simp add: len)
      show "fill_holes C bef =f (C, (bef # sss) ! 0)" using eqfI[OF _ lb(1)] by simp
      show "bef @ aft = concat (bef # sss)" using aft by simp
      fix i
      assume "i < length Cs"
      with i[OF this]
      show "?f Cs aft i =f (Cs ! i, (bef # sss) ! Suc i)" by simp
    qed        
    finally show ?case by simp
  qed auto
  then obtain sss where *: "length sss = length Cs" "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    "ss = concat sss" by blast
  have len: "length ts = length Cs" unfolding ts by auto
  assume ass: "\<And>ts sss. s = Fun f ts \<Longrightarrow>
              length ts = length Cs \<Longrightarrow>
              length sss = length Cs \<Longrightarrow> (\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)) \<Longrightarrow> ss = concat sss \<Longrightarrow> thesis"
  show thesis
    by (rule ass[OF s len *])
qed

lemma fill_holes_ctxt_main: assumes "num_holes C = Suc (length bef + length aft)"
  shows "\<exists> D. \<forall> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>"
  using assms
proof (induct C arbitrary: bef aft)
  case MHole
  show ?case
    by (rule exI[of _ \<box>], insert MHole, auto)
next
  case (MFun f cs)
  note IH = MFun(1)
  note holes = MFun(2)
  let ?p = "\<lambda> bef aft b a D cs s. map (\<lambda>i. fill_holes (cs ! i)
                        (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) (bef @ s # aft))))
               [0..<length cs] =
              b @ D\<langle>s\<rangle> # a"
  from holes IH
  have "\<exists> b D a. \<forall>s. ?p bef aft b a D cs s"
  proof (induct cs arbitrary: bef)
    case (Cons c ccs)
    have len: "length (c # ccs) = Suc (length ccs)" by simp
    show ?case 
    proof (cases "num_holes c \<le> length bef")
      case True
      hence "bef = take (num_holes c) bef @ drop (num_holes c) bef 
        \<and> length (take (num_holes c) bef) = num_holes c" by auto
      then obtain bc ba where bef: "bef = bc @ ba" and lbc: "length bc = num_holes c" by blast
      from Cons(2) have nh: "num_holes (MFun f ccs) = Suc (length ba + length aft)" unfolding bef
        by (simp add: lbc)
      from Cons(1)[OF nh Cons(3)] obtain b D a where IH: "\<And> s. ?p ba aft b a D ccs s" by auto
      show ?thesis unfolding len map_upt_Suc bef
        by (rule exI[of _ "fill_holes c bc # b"], rule exI[of _ D], rule exI[of _ a], insert IH lbc, auto)
    next
      case False
      hence "\<exists> la. num_holes c = Suc (length bef + la)" by arith
      then obtain la where nhc: "num_holes c = Suc (length bef + la)" ..
      from Cons(2) nhc have "length (take la aft) = la" by auto
      from Cons(3)[of c bef "take la aft", unfolded this, OF _ nhc]
      obtain D where D: "\<forall>s. fill_holes c (bef @ s # take la aft) = D\<langle>s\<rangle>" by auto
      show ?thesis unfolding len map_upt_Suc
        by (rule exI[of _ Nil], rule exI[of _ D], simp add: nhc D) 
    qed
  qed auto
  then obtain b D a where main: "\<And> s. ?p bef aft b a D cs s" by blast
  show ?case by (rule exI[of _ "More f b D a"], insert main, auto)
qed simp

lemma fill_holes_ctxt: assumes nh: "num_holes C = length ss"
  and i: "i < length ss"
  obtains D where "\<And> s. fill_holes C (ss[i := s]) = D \<langle> s \<rangle>"
proof -
  from id_take_nth_drop[OF i] obtain bef aft where ss: "ss = bef @ ss ! i # aft" 
    and bef: "bef = take i ss" by blast
  from bef i have bef: "length bef = i" by auto
  note len = arg_cong[OF ss, of length]
  from len nh
  have "num_holes C = Suc (length bef + length aft)" by simp
  from fill_holes_ctxt_main[OF this] obtain D where id: "\<And> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>" by blast
  {
    fix s
    have "ss[i := s] = bef @ s # aft" unfolding arg_cong[OF ss, of "\<lambda> ss. ss[i := s]"]
      using i bef by auto
    with id[of s] have "fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by simp
  }
  hence main: "\<exists> D. \<forall> s. fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by blast
  assume "\<And> D. \<lbrakk>\<And>s. fill_holes C (ss[i := s]) = D\<langle>s\<rangle>\<rbrakk> \<Longrightarrow> thesis"
  with main show thesis by blast
qed

fun split_vars :: "('f,'v)term \<Rightarrow> (('f,'v)mctxt \<times> 'v list)" where
  "split_vars (Var x) = (MHole, [x])"
| "split_vars (Fun f ts) = (MFun f (map fst (map split_vars ts)), concat (map snd (map split_vars ts)))"

lemma split_vars_num_holes: "num_holes (fst (split_vars t)) = length (snd (split_vars t))"
proof (induct t)
  case (Fun f ts)
  thus ?case by (induct ts, auto)
qed simp

lemma split_vars_vars_term_list': "snd (split_vars t) = vars_term_list' t"
proof (induct t)
  case (Fun f ts)
  thus ?case by (auto simp: vars_term_list'.simps o_def, induct ts, auto)
qed (auto simp: vars_term_list'.simps)


lemma split_vars_vars_term: "set (snd (split_vars t)) = vars_term t"
  using arg_cong[OF split_vars_vars_term_list'[of t], of set] by auto

lemma split_vars_eqf_subst: "t \<cdot> \<sigma> =f (fst (split_vars t), (map \<sigma> (snd (split_vars t))))"
proof (induct t)
  case (Fun f ts)
  have "?case = (Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ts) =f (MFun f (map (fst \<circ> split_vars) ts), concat (map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts)))"
    by (simp add: map_concat)
  also have "..." 
  proof (rule eqf_MFunI, simp, simp, unfold length_map)
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    show "map (\<lambda>t. t \<cdot> \<sigma>) ts ! i =f (map (fst \<circ> split_vars) ts ! i, map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts ! i)"
      using Fun[OF mem] i by auto
  qed
  finally show ?case by simp
qed auto

fun funas_mctxt :: "('f,'v)mctxt \<Rightarrow> 'f sig" where 
  "funas_mctxt (MFun f Cs) = insert (f, length Cs) (\<Union>(funas_mctxt ` (set Cs)))"
| "funas_mctxt _ = {}"

lemma eqf_funas_term:
  "t =f (C,ss) \<Longrightarrow> funas_term t = funas_mctxt C \<union> \<Union>(funas_term ` set ss)"
proof (induct C arbitrary: t ss)
  case (MFun f Cs t ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where
    t: "t = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and args: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    and ss: "ss = concat sss"  by auto
  let ?lhs = "\<Union> {funas_term (ts ! i) | i. i < length Cs}"
  let ?f1 = "\<lambda> i. funas_mctxt (Cs ! i)"
  let ?f2 = "\<lambda> i. \<Union>(funas_term ` set (sss ! i))"
  let ?f = "\<lambda> i. ?f1 i \<union> ?f2 i"
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    note MFun(1)[OF mem args[OF i]]
  } note IH = this
  have "funas_term t = insert (f,length Cs) ?lhs"
    unfolding t using len by (auto simp: set_conv_nth)
  also have "?lhs = \<Union> {?f i | i. i < length Cs}" using IH by (metis (hide_lams)) 
  also have "\<dots> = \<Union> {?f1 i | i. i < length Cs} \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) \<dots> = (insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs})) \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs}) = funas_mctxt (MFun f Cs)"
    by (auto simp: set_conv_nth)
  also have "\<Union> {?f2 i | i. i < length Cs} = \<Union>(funas_term ` set ss)" unfolding ss len(2)[symmetric] 
    using set_conv_nth[of sss] by auto
  finally show ?case .
next
  case MVar
  from eqfE[OF this]
  show ?case by auto
next
  case MHole
  from eqfE[OF this] show ?case by (cases ss, auto)
qed 

lemma eqf_all_ctxt_closed_step: assumes ctxt: "all_ctxt_closed F R" and 
  ass: "t =f (D,ss)" "\<And> i. i < length ts \<Longrightarrow> (ss ! i, ts ! i) \<in> R" "length ss = length ts" "funas_term t \<subseteq> F" 
  "UNION (set ts) funas_term \<subseteq> F"
  shows "(t,fill_holes D ts) \<in> R \<and> fill_holes D ts =f (D,ts)" 
proof -
  have R: "(t,fill_holes D ts) \<in> R" using ass
  proof (induct D arbitrary: t ss ts)
    case (MVar x)
    from eqfE[OF MVar(1)] have id: "t = Var x" "ss = []" by auto
    with MVar(3) have ts: "ts = []" by auto
    show ?case unfolding id ts using all_ctxt_closed_sig_reflE[OF ctxt] by auto
  next
    case MHole
    from eqfE[OF MHole(1)] have id: "ss = [t]" by (cases ss, auto)
    with MHole(3) obtain t' where ts: "ts = [t']" by (cases ts, auto)
    show ?case using MHole(2)[of 0] unfolding id ts by auto
  next
    case (MFun f Cs)
    let ?fh = "fill_holes (MFun f Cs)"
    note eqft = MFun(2)
    from MFun(5)[unfolded eqf_funas_term[OF eqft]] MFun(6)
    have F: "(f, length Cs) \<in> F" "UNION (set Cs) funas_mctxt \<subseteq> F" "UNION (set ss) funas_term \<subseteq> F" "UNION (set ts) funas_term \<subseteq> F" by auto
    from eqfE[OF eqft] have t: "t = ?fh ss" and nh: "listsum (map num_holes Cs) = length ss" by auto    
    have length_upt: "\<And> x. length [0 ..< x] = x" by auto
    show ?case unfolding t fill_holes.simps
    proof (rule all_ctxt_closedE[OF ctxt], unfold length_map length_upt, rule F(1), simp)
      let ?list = "\<lambda> Cs ts i. (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ts))"
      let ?fun = "\<lambda> Cs ts i. fill_holes (Cs ! i) (?list Cs ts i)"
      let ?val = "\<lambda> Cs ts i. map (?fun Cs ts) [0..<length Cs] ! i"
      fix i
      assume i: "i < length Cs"
      hence id: "?val Cs ts i = ?fun Cs ts i" "?val Cs ss i = ?fun Cs ss i" by auto
      from i have mem: "Cs ! i \<in> set Cs" by auto
      {
        fix j
        have "(j < length (?list Cs ts i) \<longrightarrow> (?list Cs ss i ! j, ?list Cs ts i ! j) \<in> R)
          \<and> num_holes (Cs ! i) = length (?list Cs ss i)" using MFun(3,4) nh i
        proof (induct Cs arbitrary: ts ss i j)
          case (Cons C Cs)
          let ?n = "num_holes C"
          from Cons(4) have len_ss: "length ss = ?n + listsum (map num_holes Cs)" by auto
          note len_ts = len_ss[unfolded Cons(3)] (* also for ts *)
          from len_ss len_ts have take: "length (take ?n ss) = length (take ?n ts)" by auto
          from len_ss have min_ss: "min (length ss) ?n = ?n" by auto
          from len_ts have min_ts: "min (length ts) ?n = ?n" by auto
          show ?case
          proof (cases i)
            case 0
            with Cons(2)[of j] show ?thesis by (simp add: min_ss)
          next
            case (Suc k)
            obtain ss' where ss: "ss = take ?n ss @ ss'" and ss': "ss' = drop ?n ss" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ss = drop n ss'" "ss ! (?n + i) = ss' ! i" 
                unfolding arg_cong[OF ss, of "drop (?n + n)"] 
                unfolding arg_cong[OF ss, of nth] nth_append
                unfolding drop_append ss' by (auto simp: min_ss)
            } note id_ss = this
            from len_ts obtain ts' where ts: "ts = take ?n ts @ ts'" and ts': "ts' = drop ?n ts" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ts = drop n ts'" "ts ! (?n + i) = ts' ! i"
                unfolding arg_cong[OF ts, of "drop (?n + n)"] 
                unfolding arg_cong[OF ts, of nth] nth_append
                unfolding drop_append ts' by (auto simp: min_ts)
            } note id_ts = this
            note IH = Cons(1)
            from Cons(3) have len: "length ss' = length ts'" unfolding ss' ts' by auto
            note IH = IH[OF _ this]
            from Cons(4) have nh: "listsum (map num_holes Cs) = length ss'" unfolding ss' by auto
            note IH = IH[OF _ this]
            from Cons(5) Suc have "k < length Cs" by auto
            note IH = IH[OF _ this]
            {
               fix i
               assume "i < length ts'"
               hence "?n + i < length ts" unfolding len_ts ts' nh len by auto
               from Cons(2)[OF this, unfolded id_ss id_ts]
               have "(ss' ! i, ts' ! i) \<in> R" .
            }
            note IH = IH[OF this, rule_format]
            let ?k = "listsum (map num_holes (take k Cs))"
            have "?thesis = ((j < length ts' - ?k \<and> j < num_holes (Cs ! k) \<longrightarrow>
                (drop ?k ss' ! j, drop ?k ts' ! j) \<in> R) \<and>
                num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k)))" unfolding Suc by (simp add: id_ss id_ts)
            also have "..."
            proof (intro allI conjI impI, elim conjE)
              have "num_holes (Cs ! k) = length (take (num_holes (Cs ! k)) (drop ?k ss'))" using IH by blast
              also have "\<dots> = min (length ss' - ?k) (num_holes (Cs ! k))" by simp
              finally show "num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k))" .
            next
              assume "j < length ts' - ?k" "j < num_holes (Cs ! k)"
              thus "(drop ?k ss' ! j, drop ?k ts' ! j) \<in> R"
                using IH[THEN conjunct1, THEN mp, of j] by auto
            qed
            finally show ?thesis .
          qed 
        qed auto
      }
      note main = this
      have CiF: "funas_mctxt (Cs ! i) \<subseteq> F" using F(2) i unfolding set_conv_nth by auto
      {
        have id: "?fun Cs ts i =f (Cs ! i, ?list Cs ts i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "UNION (set (?list Cs ts i)) funas_term \<subseteq> UNION (set ts) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        hence ts_list: "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" using F by auto
        have "funas_term (?fun Cs ts i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ts i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" by fact
        finally have ts: "funas_term (?fun Cs ts i) \<subseteq> F" by blast
        have id: "?fun Cs ss i =f (Cs ! i, ?list Cs ss i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "funas_term (?fun Cs ss i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ss i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ss i)) funas_term \<subseteq> UNION (set ss) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        also have "\<dots> \<subseteq> F" using F by auto
        finally have ss: "funas_term (?fun Cs ss i) \<subseteq> F" by blast
        note ts ss ts_list
      } note F_main = this
      have "((?val Cs ss i, ?val Cs ts i) \<in> R) = ((?fun Cs ss i, ?fun Cs ts i) \<in> R)" unfolding id ..
      also have "..." 
        by (rule MFun(1)[OF mem eqfI[OF refl] _ _ F_main(2-3)], insert main MFun(4), auto)
      finally show "(?val Cs ss i, ?val Cs ts i) \<in> R" .
      show "funas_term (?val Cs ss i) \<subseteq> F" unfolding id by fact
      show "funas_term (?val Cs ts i) \<subseteq> F" unfolding id by fact
    qed 
  qed
  show ?thesis
    by (rule conjI[OF R eqfI], insert eqfE[OF ass(1)] ass(3), auto)
qed

fun map_mctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)mctxt \<Rightarrow> ('g,'v)mctxt" where
  "map_mctxt _ (MVar x) = (MVar x)"
| "map_mctxt _ (MHole) = MHole"
| "map_mctxt fg (MFun f Cs) = MFun (fg f) (map (map_mctxt fg) Cs)"

fun ground_mctxt :: "('f,'v)mctxt \<Rightarrow> bool" where 
  "ground_mctxt (MVar _) = False"
| "ground_mctxt MHole = True"
| "ground_mctxt (MFun f Cs) = Ball (set Cs) ground_mctxt"

lemma split_vars_ground: "split_vars t = (C,xs) \<Longrightarrow> ground_mctxt C"
proof (induct t arbitrary: C xs)
  case (Fun f ts C xs)
  from Fun(2)[simplified] obtain Cs where C: "C = MFun f Cs" and Cs: "Cs = map (fst \<circ> split_vars) ts" by auto
  show ?case unfolding C ground_mctxt.simps
  proof
    fix C
    assume "C \<in> set Cs"
    from this[unfolded Cs] obtain t where t: "t \<in> set ts" and C: "C = fst (split_vars t)" unfolding o_def by auto
    from C obtain xs where split: "split_vars t = (C,xs)" by (cases "split_vars t", auto)
    show "ground_mctxt C"
      by (rule Fun(1)[OF t split]) 
  qed
qed auto

lemma split_vars_ground_vars: assumes "ground_mctxt C" and "num_holes C = length xs" 
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
  using assms
proof (induct C arbitrary: xs)
  case (MHole xs)
  thus ?case by (cases xs, auto)
next
  case (MFun f Cs xs)
  have "fill_holes (MFun f Cs) (map Var xs) =f (MFun f Cs, map Var xs)"
    by (rule eqfI, insert MFun(3), auto)
  from eqf_MFunE[OF this] 
  obtain ts xss where fh: "fill_holes (MFun f Cs) (map Var xs) = Fun f ts"
    and lent: "length ts = length Cs"
    and lenx: "length xss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, xss ! i)"
    and id: "map Var xs = concat xss" by auto
  from arg_cong[OF id, of "map the_Var"] have id2: "xs = concat (map (map the_Var) xss)" 
    by (auto simp: o_def map_concat)
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    with MFun(2) have ground: "ground_mctxt (Cs ! i)" by auto
    have "map Var (map the_Var (xss ! i)) = map id (xss ! i)" unfolding map_map o_def map_eq_conv
    proof
      fix x
      assume "x \<in> set (xss ! i)"
      with lenx i have "x \<in> set (concat xss)" by auto
      from this[unfolded id[symmetric]] show "Var (the_Var x) = id x" by auto
    qed
    hence idxss: "map Var (map the_Var (xss ! i)) = xss ! i" by auto
    note rec = eqfE[OF args[OF i]]
    note IH = MFun(1)[OF mem ground, of "map the_Var (xss ! i)", unfolded rec(2) idxss rec(1)[symmetric]]
    from IH have "split_vars (ts ! i) = (Cs ! i, map the_Var (xss ! i))" by auto
    note this idxss
  }
  note IH = this
  have "?case = (map fst (map split_vars ts) = Cs \<and> concat (map snd (map split_vars ts)) = concat (map (map the_Var) xss))"
    unfolding fh unfolding id2 by auto
  also have "\<dots>"
  proof (rule conjI[OF nth_equalityI[OF _ allI[OF impI]] arg_cong[of _ _ concat, OF nth_equalityI, rule_format]], unfold length_map lent lenx)
    fix i
    assume i: "i < length Cs" 
    with arg_cong[OF IH(2)[OF this], of "map the_Var"]
      IH[OF this] show "map snd (map split_vars ts) ! i = map (map the_Var) xss ! i" using lent lenx by auto
  qed (insert IH lent, auto)
  finally show ?case .
qed auto

lemma ground_map_mctxt[simp]: "ground_mctxt (map_mctxt fg C) = ground_mctxt C"
  by (induct C, auto)

lemma num_holes_map_mctxt[simp]: "num_holes (map_mctxt fg C) = num_holes C"
proof (induct C)
  case (MFun f Cs)
  thus ?case by (induct Cs, auto)
qed auto

lemma split_vars_map_mctxt: assumes split: "split_vars t = (map_mctxt fg C, xs)"
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
proof -
  from split_vars_ground[OF split] have ground: "ground_mctxt C" by simp
  from split_vars_num_holes[of t, unfolded split] have nh: "num_holes C = length xs" by auto
  show ?thesis
    by (rule split_vars_ground_vars[OF ground nh])
qed

lemma subst_eq_map_decomp: assumes "t \<cdot> \<sigma> = map_term fg s"
  shows "\<exists> C xs \<delta>s. s =f (C,\<delta>s) \<and> split_vars t = (map_mctxt fg C, xs) \<and> (\<forall> i < length xs. 
    \<sigma> (xs ! i) = map_term fg (\<delta>s ! i))" 
  using assms
proof (induct t arbitrary: s)
  case (Var x s)
  show ?case
    by (rule exI[of _ MHole], rule exI[of _ "[x]"], rule exI[of _ "[s]"], insert Var, auto)
next
  case (Fun g ts s)
  from Fun(2) obtain f ss where s: "s = Fun f ss" and g: "g = fg f" by (cases s, auto)
  from Fun(2)[unfolded s] have id: "map (\<lambda> t. t \<cdot> \<sigma>) ts = map (map_term fg) ss" by auto
  from arg_cong[OF this, of length] have len: "length ts = length ss" by auto
  from map_nth_conv[OF id] have args: "\<And> i. i < length ts \<Longrightarrow> ts ! i \<cdot> \<sigma> = map_term fg (ss ! i)" by auto
  let ?P = "\<lambda> C xs \<delta>s i. ss ! i =f (C, \<delta>s) \<and>
          split_vars (ts ! i) = (map_mctxt fg C, xs) \<and>
          (\<forall>i<length xs. \<sigma> (xs ! i) = map_term fg (\<delta>s ! i))"
  {
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    note IH = Fun(1)[OF this args[OF i]]
  }
  hence "\<forall> i. \<exists> C xs \<delta>s. i < length ts \<longrightarrow> ?P C xs \<delta>s i" by blast
  from choice[OF this] obtain Cs where "\<forall> i. \<exists> xs \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) xs \<delta>s i" by blast
  from choice[OF this] obtain xss where "\<forall> i. \<exists> \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) (xss i) \<delta>s i" by blast
  from choice[OF this] obtain \<delta>ss where IH: "\<And> i. i < length ts \<Longrightarrow> ?P (Cs i) (xss i) (\<delta>ss i) i" by blast
  let ?n = "[0 ..< length ts]"
  let ?Cs = "map Cs ?n"
  let ?C = "MFun f ?Cs"
  let ?xs = "concat (map xss ?n)"
  let ?\<delta>s = "concat (map \<delta>ss ?n)"
  let ?g = "fg f"
  show ?case unfolding s g 
  proof (rule exI[of _ ?C], rule exI[of _ ?xs], rule exI[of _ ?\<delta>s], intro conjI)
    show "Fun f ss =f (?C, ?\<delta>s)" 
      by (rule eqf_MFunI, insert IH len, auto)
  next
    have "(split_vars (Fun ?g ts) = (map_mctxt fg ?C, ?xs)) = (map (fst \<circ> split_vars) ts = map (map_mctxt fg \<circ> Cs) [0..<length ss] \<and>
      concat (map (snd \<circ> split_vars) ts) = ?xs)" 
      (is "?goal = _") unfolding split_vars.simps map_mctxt.simps
      using len by auto
    also have "\<dots>"
      by (rule conjI[OF nth_map_conv arg_cong[of _ _ concat, OF nth_equalityI]], insert IH len, auto)
    finally show ?goal .
  next
    show "\<forall> i < length ?xs. \<sigma> (?xs ! i) = map_term fg (?\<delta>s ! i)"
    proof (rule concat_all_nth, simp, unfold length_map length_upt)
      fix i
      assume "i < length ts - 0"
      hence i: "i < length ts" by auto
      from IH[OF i] have "split_vars (ts ! i) = (map_mctxt fg (Cs i), xss i)" by blast
      from split_vars_map_mctxt[OF this] split_vars_num_holes[of "fill_holes (Cs i) (map Var (xss i))"]
      have len: "length (xss i) = num_holes (Cs i)" by simp
      also have "\<dots> = length (\<delta>ss i)" by (rule eqfE(2), insert IH[OF i], auto)
      finally 
      show "length (map xss [0..<length ts] ! i) = length (map \<delta>ss [0..<length ts] ! i)" using i by auto
    qed (insert IH, auto)
  qed
qed

lemma map_term_fill_holes: "num_holes C = length ss \<Longrightarrow> map_term fg (fill_holes C ss) =f (map_mctxt fg C, map (map_term fg) ss)"
proof (induct C arbitrary: ss)
  case (MHole ss)
  thus ?case by (cases ss, auto)
next
  case MVar thus ?case by auto
next
  case (MFun f Cs ss)
  from MFun(2) have "fill_holes (MFun f Cs) ss =f (MFun f Cs, ss)" by auto
  from eqf_MFunE[OF this] obtain ts sss where fh: "fill_holes (MFun f Cs) ss = Fun f ts"
    and lts: "length ts = length Cs"
    and lsss: "length sss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    and sss: "ss = concat sss" by auto
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    from MFun(1)[OF mem]  eqfE[OF args[OF i]] have 
      "map_term fg (ts ! i) =f (map_mctxt fg (Cs ! i), map (map_term fg) (sss ! i))" by auto
  } note IH = this
  show ?case unfolding fh 
    unfolding map_mctxt.simps sss map_concat map_term.simps
  proof (rule eqf_MFunI, unfold length_map, simp add: lsss, simp add: lts)
    fix i
    assume i: "i < length Cs"
    have "(map (map_term fg) ts ! i =f (map (map_mctxt fg) Cs ! i, map (map (map_term fg)) sss ! i)) =
      (map_term fg (ts ! i) =f (map_mctxt fg (Cs ! i), map (map_term fg) (sss ! i)))" (is "?goal = _")
      using i lts lsss by auto
    also have "\<dots>" by (rule IH[OF i])
    finally show ?goal .
  qed
qed

fun poss_mctxt :: "('f,'v)mctxt \<Rightarrow> pos set" where
  "poss_mctxt (MVar x) = {\<epsilon>}"
| "poss_mctxt MHole = {}"
| "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` poss_mctxt (cs ! i)) [0 ..< length cs]))"

lemma poss_mctxt_simp[simp]: "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> {i <# p | i p. i < length cs \<and> p \<in> poss_mctxt (cs ! i)}"
  by auto
declare poss_mctxt.simps(3)[simp del]

fun hole_poss :: "('f,'v)mctxt \<Rightarrow> pos set" where
  "hole_poss (MVar x) = {}"
| "hole_poss MHole = {\<epsilon>}"
| "hole_poss (MFun f cs) = \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` hole_poss (cs ! i)) [0 ..< length cs]))"

lemma hole_poss_simp[simp]: "hole_poss (MFun f cs) = {i <# p | i p. i < length cs \<and> p \<in> hole_poss (cs ! i)}"
  by auto
declare hole_poss.simps(3)[simp del]

fun fill_holes_mctxt :: "('f,'v)mctxt \<Rightarrow> ('f,'v)mctxt list \<Rightarrow> ('f,'v)mctxt" where
  "fill_holes_mctxt (MVar x) _ = MVar x"
| "fill_holes_mctxt MHole [t] = t"
| "fill_holes_mctxt (MFun f cs) ts = (MFun f (map (\<lambda> i. fill_holes_mctxt (cs ! i) 
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

lemma fill_holes_mctxt_MFun: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes_mctxt (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes_mctxt (MFun f Cs) (concat ss) = MFun f ts" 
  unfolding fill_holes_mctxt.simps mctxt.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

end

