(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Position 
imports 
  Main 
  Show_Instances
begin

datatype pos = Empty ("\<epsilon>") | PCons "nat" "pos" (infixr "<#" 70)

primrec append :: "pos \<Rightarrow> pos \<Rightarrow> pos" (infixr "<#>" 70)
where append_Empty: "\<epsilon><#>q = q"
    | append_PCons:  "(i<#p)<#>q = i<#(p<#>q)"

abbreviation singleton :: "nat \<Rightarrow> pos" ("<_>") where "<i> == i<#\<epsilon>"
abbreviation snoc :: "pos \<Rightarrow> nat \<Rightarrow> pos" (infixl "#>" 70) where "p#>i == p <#> <i>"

lemma not_PCons_self[simp]: "p \<noteq> i<#p" by (induct p) auto
lemmas not_PCons_self2[simp] = not_PCons_self[symmetric]
lemma not_snoc_self[simp]: "p \<noteq> p#>i" by (induct p) auto
lemmas not_snoc_self2[simp] = not_snoc_self[symmetric]
lemma impossible_PCons: "size p \<le> size q \<Longrightarrow> p = i<#q = False" by (induct p) auto
lemma size_0_conv[iff]: "(0 = size p) = (p = \<epsilon>)" by (cases p) auto
lemma size_greater_0_conf[iff]: "(0 < size p) = (p \<noteq> \<epsilon>)" by (cases p) auto 

lemma append_assoc[simp]: fixes p q r :: "pos" shows "(p <#> q) <#> r = p <#> (q <#> r)" by (induct p) auto
lemma append_Empty2[simp]: "p <#> \<epsilon> = p" by (induct p) auto
lemma size_append[iff]: fixes p :: "pos" shows "size(p <#> q) = size p + size q" by (induct p) auto

interpretation semigroup_append: semigroup_add "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by unfold_locales simp
interpretation monoid_append: monoid_add "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" "\<epsilon>" by unfold_locales simp+

lemma append_is_Empty_conv[iff]: "(p <#> q = \<epsilon>) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma Empty_is_append_conv[iff]: "(\<epsilon> = p <#> q) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma append_self_conv[iff]: "(p <#> q = p) = (q = \<epsilon>)" by (induct p) auto
lemma self_append_conv[iff]: "(p = p <#> q) = (q = \<epsilon>)" by (induct p) auto

lemma append_eq_append_conv[simp,no_atp]:
 fixes p q r s :: "pos"
 assumes "size p = size q \<or> size r = size s"
 shows "(p <#> r = q <#> s) = (p = q \<and> r = s)"
using assms proof (induct p arbitrary: q)
 case Empty thus ?case by (cases q,simp,force)
next
 case (PCons i p) thus ?case by (cases q,force,simp)
qed

lemma same_append_eq[iff]: fixes p :: "pos" shows "(p <#> q = p <#> r) = (q = r)" by simp
lemma same_append_eq'[iff]: fixes p :: "pos" shows "(q <#> p = r <#> p) = (q = r)" by simp

primrec rev :: "pos \<Rightarrow> pos"
where "rev \<epsilon> = \<epsilon>"
    | "rev(i<#p) = rev p #> i"

lemma rev_append[simp]: fixes p :: "pos" shows "rev(p <#> q) = rev q <#> rev p" by (induct p) auto
lemma rev_rev_ident[simp]: fixes p :: "pos" shows "rev(rev p) = p" by (induct p) auto
lemma rev_size[simp]: fixes p :: "pos" shows "size(rev p) = size p" by (induct p) auto
lemma rev_inj: assumes "rev xs = rev ys" shows "xs = ys" using arg_cong[OF assms, of rev] by simp

lemma rev_induct[case_names Empty snoc]:
 fixes P :: "pos \<Rightarrow> bool" assumes A: "P \<epsilon>" and B: "\<And>i q. P q \<Longrightarrow> P(q#>i)" 
 shows "P p"
using assms by (simplesubst rev_rev_ident[symmetric],induct rule: pos.induct[where pos = "rev p"]) simp_all

lemma [simp]: "i<#p \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> i<#p" by simp

lemma [simp]: "p#>i \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> p#>i" by simp

instantiation pos :: order begin
 definition less_eq_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_eq_pos p q == (\<exists>r. (p <#> r) = q)"
 definition less_pos    :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_pos p q    == (p \<le> q \<and> p \<noteq> q)"
 
 instance proof
  fix p q :: pos show "(p < q) = (p \<le> q \<and> \<not>(q \<le> p))" unfolding less_eq_pos_def less_pos_def by auto
 next
  fix p :: pos show "(p \<le> p)" unfolding less_eq_pos_def by auto
 next
  fix p q r :: pos assume "p \<le> q" and "q \<le> r" thus "p \<le> r" unfolding less_eq_pos_def by auto
 next
  fix p q :: pos assume "p \<le> q" and "q \<le> p"
  from `p \<le> q` obtain r where "p <#> r = q" unfolding less_eq_pos_def by best
  from `q \<le> p` obtain s where "q <#> s = p" unfolding less_eq_pos_def by best
  hence "p <#> (r <#> s) = p" unfolding `p <#> r = q`[symmetric] by auto
  hence "r <#> s = \<epsilon>" by simp 
  hence "r = \<epsilon>" and "s = \<epsilon>" by auto
  from `p <#> r = q` show "p = q" unfolding `r = \<epsilon>` by simp
qed
end

lemma epsilon_least[simp]: "(\<epsilon> :: pos) \<le> p" unfolding less_eq_pos_def by auto
lemma less_eq_simps[simp]: 
  "p \<le> p <#> q" 
  "(p <#> q1 \<le> p <#> q2) = (q1 \<le> q2)" 
  "(i <# q1 \<le> \<epsilon>) = False"
  "(i <# q1 \<le> j <# q2) = (i = j \<and> q1 \<le> q2)"
  "(p <#> q \<le> p) = (q = \<epsilon>)" unfolding less_eq_pos_def by auto

lemma [code]: "(\<epsilon> :: pos) \<le> p = True" 
  "(i <# q1 \<le> \<epsilon>) = False"
  "(i <# q1 \<le> j <# q2) = (i = j \<and> q1 \<le> q2)" by auto

lemma less_simps[simp]: 
  "(p < p <#> q) = (q \<noteq> \<epsilon>)" 
  "(p <#> q1 < p <#> q2) = (q1 < q2)" 
  "(p < \<epsilon>) = False"
  "(i <# q1 < j <# q2) = (i = j \<and> q1 < q2)"
  "(p <#> q < p) = False" unfolding less_pos_def less_eq_pos_def by auto

lemma prefix_smaller[simp]: fixes p :: "pos" assumes "p < q" shows "size p < size q"
using assms unfolding less_pos_def less_eq_pos_def by auto

lemma size_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q. size q < size p \<longrightarrow> P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) iprover

lemma prefix_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q<p. P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) auto

interpretation pos: semigroup_mult "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by (unfold_locales) simp

interpretation pos: monoid_mult "\<epsilon> :: pos" "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by unfold_locales simp_all

instantiation pos :: one begin
  definition one_pos_def[simp]: "1 = \<epsilon>"
  instance by (intro_classes)
end

instantiation pos :: times begin
  definition times_pos_def[simp]: "times p q = p <#> q"
  instance by (intro_classes)
end

instantiation pos :: power begin
  instance by (intro_classes)
end

instantiation pos :: semigroup_mult begin
  instance proof
    fix a b c::pos show "a * b * c = a * (b * c)" by simp
  qed
end

lemma power_append_distr: "p^(m+n) = p^(m::nat) <#> p^n" by (induct m) auto

lemma power_pos_Suc: "p^(Suc n) = p^n <#> p"
proof -
  have "p ^ Suc n = p ^ (n + Suc 0)" by simp
  also have "... = p ^ n <#> p" unfolding power_append_distr by auto 
  finally show ?thesis .
qed

lemma power_subtract_less_eq: "(p::pos)^((n::nat) - m) \<le> p^n"
proof(cases "m \<ge> n",simp)
 case False
  hence "(n -m) + m = n" by auto
  thus ?thesis unfolding less_eq_pos_def using power_append_distr by metis
qed


lemma power_size: fixes p :: "pos" shows "size (p^n) = size p * n" 
 by (induct n, simp, auto)

primrec pos_prefix :: "pos \<Rightarrow> pos \<Rightarrow> pos option"
where pre_Empty: "pos_prefix \<epsilon> q = Some q"
    | pre_PCons: "pos_prefix (i <# p) q = 
     (case q of \<epsilon> \<Rightarrow> None | j <# q' \<Rightarrow> if i=j then pos_prefix p q' else None)" 

lemma prefix_sound:
 "pos_prefix p q = Some r \<Longrightarrow> p <#> r = q"
proof (induct p arbitrary: q,simp)
 case (PCons i p) note IH = this
  thus ?case
   proof(cases q,simp)
    case (PCons j q')
     with IH have pq':"pos_prefix p q' = Some r" using pre_PCons option.distinct(1) by (cases "i = j",simp,force)
     have "i = j" using PCons option.distinct(1) pre_PCons IH by (cases "i = j",simp,force) 
     with pq' IH PCons show "(i <# p) <#> r = q" by auto
  qed
qed

lemma prefix_exists: 
 "p \<le> q \<Longrightarrow> \<exists> q'. p <#> q' = q \<and> pos_prefix p q = Some q'"
proof (induct p arbitrary: q,simp) 
 case (PCons i p') note IH = this
 from PCons have "\<exists> q'. q = i <# q'" unfolding less_eq_pos_def by auto
 from this obtain q' where q_is: "q = i <# q'" and leq': "p' \<le> q'" using PCons by auto
 hence res: "pos_prefix (i <# p') q = pos_prefix p' q'" unfolding pos_prefix_def by auto
 hence "\<exists> r. p' <#> r = q' \<and> pos_prefix p' q' = Some r" using leq' IH by auto
 from this obtain r where "p' <#> r = q'" and "pos_prefix p' q' = Some r" by auto
 thus ?case using q_is res by auto 
qed

fun pos_suffix :: "pos \<Rightarrow> pos \<Rightarrow> pos option"
where "pos_suffix p q = (case pos_prefix (rev p) (rev q) of
 None \<Rightarrow> None | Some q' \<Rightarrow> Some (rev q'))" 

lemma suffix_sound: "pos_suffix p q = Some r \<Longrightarrow> r <#> p = q" (is "?l \<Longrightarrow> ?r")
proof-
 assume ?l
 hence "pos_prefix (rev p) (rev q) = Some (rev r)" using pos_suffix_def option.distinct(1) 
  by (cases "pos_prefix (rev p) (rev q)",simp,auto)
 thus ?thesis using prefix_sound rev_append rev_inj by auto
qed

lemma suffix_complete: 
 "\<exists>r. r <#> p = q \<Longrightarrow> \<exists>r. r <#> p = q \<and> pos_suffix p q = Some r" (is "?ls \<Longrightarrow> ?rs")
proof-
 assume ?ls
 hence "rev p \<le> rev q" using rev_append less_eq_pos_def by auto
 hence "\<exists> q'. (rev p) <#> q' = (rev q) \<and> pos_prefix (rev p) (rev q) = Some q'" using prefix_exists by auto
 then obtain r where r:"(rev p) <#> r = (rev q)" and "pos_prefix (rev p) (rev q) = Some r" by auto
 hence "pos_suffix p q = Some (rev r)" and "rev ((rev r) <#> p) = rev q" using rev_rev_ident rev_append by auto
 thus ?rs using rev_inj by auto
qed

text {* Hide unqualified names of constants that are already used elsewhere. *}
hide_const (open)  append rev

lemma pos_exhaust_rev[case_names empty snoc]:
  assumes "p = \<epsilon> \<Longrightarrow> P" and "\<And>\<pi> i. p = \<pi>#>i \<Longrightarrow> P" shows "P"
using assms by (induct p rule:rev_induct) auto

lemma epsilon_power_epsilon[simp]: "\<epsilon> ^ n = \<epsilon>" by (induct n, auto)

fun parallel_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" (infixr "\<bottom>" 64)
  where "\<epsilon> \<bottom> _ = False"
      | "_ \<bottom> \<epsilon> = False"
      | "i <# p \<bottom> j <# q = (i \<noteq> j \<or> p \<bottom> q)"

lemma parallel_pos: "p \<bottom> q = (\<not> p \<le> q \<and> \<not> q \<le> p)"
proof (induct p arbitrary: q)
  case Empty
  show ?case by simp
next
  case (PCons i p q) note IH = this
  show ?case
  proof (cases q)
    case Empty
    thus ?thesis by simp
  next
    case (PCons j q')
    show ?thesis 
    proof (cases "i = j")
      case False
      thus ?thesis unfolding PCons by simp
    next
      case True
      with IH[of q'] show ?thesis unfolding PCons by auto
    qed
  qed
qed

lemma parallel_pos_prefix: "p1 \<bottom> p2 \<Longrightarrow>
  \<exists> p i j q1 q2. p1 = p <#> i <# q1 \<and> p2 = p <#> j <# q2 \<and> i \<noteq> j"
proof (induct p1 arbitrary: p2)
  case Empty thus ?case by simp
next
  case (PCons i p1)
  from PCons(2) obtain j p2' where p2: "p2 = j <# p2'" by (cases p2, auto)
  with PCons(2) have "i \<noteq> j \<or> (i = j) \<and> p1 \<bottom> p2'" by simp
  thus ?case
  proof
    assume ij: "i \<noteq> j"
    show ?case
    proof (intro exI conjI)
      show "i <# p1 = \<epsilon> <#> i <# p1" by simp
    next
      show "p2 = \<epsilon> <#> j <# p2'" unfolding p2 by simp
    qed (rule ij)
  next
    assume "i = j \<and> p1 \<bottom> p2'" hence
      p12: "p1 \<bottom> p2'" and ij: "i = j" by auto
    from PCons(1)[OF p12] obtain p i' j' q1 q2 where p1: "p1 = p <#> i' <# q1"
      and p2': "p2' = p <#> j' <# q2" and ij': "i' \<noteq> j'" by blast
    show ?thesis
    proof (intro exI conjI)
      show "i <# p1 = (i <# p) <#> i' <# q1" unfolding p1 by simp
    next
      show "p2 = (i <# p) <#> j' <# q2" unfolding p2 p2' ij by simp
    qed (rule ij')
  qed
qed

lemma position_cases: "p \<le> q \<or> q < p \<or> p \<bottom> q"
proof (cases "p \<bottom> q")
  case True thus ?thesis by auto
next
  case False
  from this[unfolded parallel_pos] have "p \<le> q \<or> (\<not> p \<le> q \<and> q \<le> p)" by auto
  thus ?thesis
  proof
    assume "p \<le> q" thus ?thesis by auto
  next
    assume "\<not> p \<le> q \<and> q \<le> p" 
    hence "q < p" by auto
    thus ?thesis by auto
  qed
qed

lemma parallel_pos_sym: "p1 \<bottom> p2 \<Longrightarrow> p2 \<bottom> p1"
  unfolding parallel_pos by auto

lemma less_pos_def': "(p < q) = (\<exists> r. q = p <#> r \<and> r \<noteq> \<epsilon>)" (is "?l = ?r")
proof
  assume ?r then obtain r where id: "p <#> r = q" "r \<noteq> \<epsilon>" by blast
  show ?l unfolding less_pos_def less_eq_pos_def
    by (rule conjI[OF exI[of _ r]], unfold id[symmetric], insert id(2), auto)
next
  assume ?l
  then obtain r where id: "p <#> r = q" "p \<noteq> q" unfolding less_pos_def less_eq_pos_def by blast
  show ?r
    by (rule exI[of _ r], insert id(2), unfold id[symmetric], auto)
qed    

lemma pos_append_cases: 
  "p1 <#> p2 = q1 <#> q2 \<Longrightarrow> (\<exists> q3. p1 = q1 <#> q3 \<and> q2 = q3 <#> p2) \<or> 
      (\<exists> p3. q1 = p1 <#> p3 \<and> p2 = p3 <#> q2)"
proof (induct p1 arbitrary: q1)
  case Empty
  thus ?case by auto
next
  case (PCons i p1' q1) note IH = this
  show ?case 
  proof (cases q1)
    case Empty
    thus ?thesis using IH(2) by auto
  next
    case (PCons j q1')
    with IH(2) have id: "p1' <#> p2 = q1' <#> q2" and ij: "i = j" by auto
    from IH(1)[OF id]
    show ?thesis unfolding PCons ij by auto
  qed
qed

lemma pos_less_eq_append_not_parallel:
 assumes "q \<le> p <#> q'"
 shows "\<not> (q \<bottom> p)"
proof-
  from assms obtain r where "q <#> r = p <#> q'" unfolding less_eq_pos_def ..  
  hence dec:"(\<exists> q3. q = p <#> q3 \<and> q' = q3 <#> r) \<or> 
   (\<exists> p3. p = q <#> p3 \<and> r = p3 <#> q')" (is "?a \<or> ?b") by (rule pos_append_cases)
  hence "p \<le> q \<or> q \<le> p" unfolding less_eq_pos_def by blast
  thus ?thesis unfolding parallel_pos by auto
qed

lemma less_pos_power_split: "q < p^m \<Longrightarrow> \<exists> p' k. q = p^k <#> p' \<and> p' < p \<and> k < m"
proof (induct m arbitrary: q)
  case 0
  thus ?case by auto
next
  case (Suc n q)
  show ?case
  proof (cases "q < p")
    case True
    show ?thesis
      by (rule exI[of _ q], rule exI[of _ 0], insert True, auto)
  next
    case False
    from Suc(2) obtain r where pn: "p <#> p^n = q <#> r" unfolding less_pos_def' by auto
    from pos_append_cases[OF this]
    have "\<exists> r. q = p <#> r"
    proof
      assume "\<exists> s. p = q <#> s \<and> r = s <#> p ^ n"
      then obtain s where p: "p = q <#> s" by auto
      with False show ?thesis by auto
    qed auto
    then obtain r where q: "q = p <#> r" by auto
    with Suc(2) have "r < p ^ n" by simp
    from Suc(1)[OF this] obtain p' k where r: "r = p ^ k <#> p'" "p' < p" "k < n" by auto
    show ?thesis unfolding q
      by (rule exI[of _ p'], rule exI[of _ "Suc k"], insert r, auto)
  qed
qed

instantiation pos :: "show" begin
  fun
    shows_pos :: "pos \<Rightarrow> shows"
  where
    "shows_pos (i <# \<epsilon>) = shows (Suc i)" |
    "shows_pos (i <# ps) = (shows (Suc i) +@+ shows_pos ps)" |
    "shows_pos \<epsilon> = shows_string ''epsilon''"
  definition "shows_prec (d::nat) (p::pos) = shows_pos p"

  lemma assoc_pos: "shows_prec d (p::pos) r @ s = shows_prec d p (r @ s)"
  unfolding shows_prec_pos_def
  proof (induct p)
    case (PCons i ps) thus ?case by (cases ps) simp_all
  qed simp
  standard_shows_list assoc_pos
end

fun proper_prefix_list :: "pos \<Rightarrow> pos list" where
  "proper_prefix_list \<epsilon> = []"
| "proper_prefix_list (PCons i p) = \<epsilon> # map (PCons i) (proper_prefix_list p)"

lemma proper_prefix_list[simp]: "set (proper_prefix_list p) = {q. q < p}"
proof (induction p)
  case Empty
  show ?case by auto
next
  case (PCons i p)
  note IH = this
  show ?case (is "?l = ?r")
  proof (rule set_eqI)
    fix q
    show "q \<in> ?l = (q \<in> ?r)"
    proof (cases q)
      case Empty
      have less: "\<epsilon> < i <# p" unfolding less_pos_def by auto
      show ?thesis unfolding Empty using less by auto
    next
      case (PCons j q')
      show ?thesis unfolding PCons by (auto simp: IH)
    qed
  qed
qed

definition prefix_list :: "pos \<Rightarrow> pos list" where
  "prefix_list p \<equiv> p # proper_prefix_list p"

lemma prefix_list[simp]: "set (prefix_list p) = { q. q \<le> p}"
  unfolding prefix_list_def by auto

definition bounded_postfixes :: "pos \<Rightarrow> pos list \<Rightarrow> pos list" where
  "bounded_postfixes p ps \<equiv> map the [opt\<leftarrow>map (pos_prefix p) ps . opt \<noteq> None]"

lemma bounded_postfixes[simp]: "set (bounded_postfixes p ps) = { r. p <#> r \<in> set ps}" (is "?l = ?r")
proof -
  note d = bounded_postfixes_def
  {
    fix r
    assume "r \<in> ?l"
    from this[unfolded d] obtain q res where q: "q \<in> set ps" and pre: "pos_prefix p q = res" and 
      res: "res \<noteq> None" and r: "r = the res" by auto
    from res r have res: "res = Some r" by (cases res, auto)
    from prefix_sound[OF pre[unfolded res]] q have "r \<in> ?r" by auto
  }
  moreover
  {
    fix r
    assume "r \<in> ?r"
    then obtain q where q: "q \<in> set ps" and p: "p <#> r = q" by auto
    hence "p \<le> q" by auto
    from prefix_exists[OF this] obtain q' where "p <#> q' = q \<and> pos_prefix p q = Some q'" by auto
    with p have "pos_prefix p q = Some r" by auto
    with `q \<in> set ps` have "r \<in> ?l" unfolding d by force
  }
  ultimately show ?thesis by blast
qed

primrec to_list :: "pos \<Rightarrow> nat list"
where "to_list \<epsilon> = []"
    | "to_list (i<#p) = i # (to_list p)"

end

