(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory String_Reversal_Impl
imports
  Trs_Impl
  String_Reversal
  QDP_Framework_Impl
begin

definition
  check_unary_signature :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_unary_signature R \<equiv>
     check_all (\<lambda>(l, r). unary_term l \<and> unary_term r) R
       <+? (\<lambda>r. ''the rule '' +#+ shows_nl +@+ shows_rule r +@+ shows_nl
        +@+ ''violates the condition that all function symbols'' +#+ shows_nl
        +@+ shows ''have to be unary'')"
  
lemma check_unary_signature_sound[simp]: "isOK(check_unary_signature R) = unary_sig (funas_trs (set R))" (is "_ = ?r")
proof (unfold check_unary_signature_def, simp)
  show "(\<forall> x \<in> set R. (\<lambda>(l,r). unary_term l \<and> unary_term r) x) = ?r" (is "?l = _")
  proof
    assume ?l
    show ?r unfolding unary_sig_def funas_trs_def funas_rule_def [abs_def]
    proof (rule ballI2)
      fix f i
      assume "(f,i) \<in> (\<Union>r \<in> set R. funas_term (fst r) \<union> funas_term (snd r))"
      then obtain r where r: "r \<in> set R" and lr: "(f,i) \<in> funas_term (fst r) \<union> funas_term (snd r)" by auto
      from lr show "i = 1"
      proof
        assume "(f,i) \<in> funas_term (fst r)"
        thus "i = 1" using `?l` unary_funas_conv[of "fst r" f i] r by auto
      next
        assume "(f,i) \<in> funas_term (snd r)"
        thus "i = 1" using `?l` unary_funas_conv[of "snd r" f i] r by auto
      qed
    qed
  next
    assume ?r
    show ?l (is "\<forall> x \<in> set R. ?P x")
    proof
      fix l r
      assume lr: "(l, r) \<in> set R"
      with `?r` lr have l: "\<forall> (f,i) \<in> funas_term l. i = 1" and r: "\<forall> (f,i) \<in> funas_term r. i = 1"
        unfolding unary_sig_def funas_trs_def funas_rule_def [abs_def] by auto
      show "unary_term l \<and> unary_term r" using funas_unary_conv[OF l] funas_unary_conv[OF r] ..
    qed
  qed
qed

definition
  string_reversal_tt :: "('tp, 'f::show, 'v::show) tp_ops \<Rightarrow> 'tp proc"
where
  "string_reversal_tt I trs \<equiv> (let
    rs = tp_ops.rules I trs;
    r  = tp_ops.R I trs;
    s  = tp_ops.Rw I trs
  in check_return (do {
     check_unary_signature rs
  }) (tp_ops.mk I False [] (map rev_rule r) (map rev_rule s)))"
(*FIXME: default value for normal-form substitutions*)

context tp_spec
begin
lemma string_reversal_tt:
  "sound_tt_impl (string_reversal_tt I)"
proof 
  fix tp tp'
  assume ok: "string_reversal_tt I tp = return tp'"
    and fin: "SN_qrel (qreltrs tp')"
  let ?Q = "set (Q tp)"
  let ?R = "set (R tp)"
  let ?S = "set (Rw tp)"
  let ?RS = "?R \<union> ?S"
  from ok[unfolded string_reversal_tt_def Let_def, simplified]
  have unary: "unary_sig (funas_trs ?RS)" 
    and tp': "tp' = mk False [] (map rev_rule (R tp)) (map rev_rule (Rw tp))" 
    by auto
    (*FIXME: default value for normal-form substitutions*)
  from fin have "SN_qrel (False, {}, rev_trs ?R, rev_trs ?S)" unfolding tp'
    unfolding mk_sound
    by (auto simp: rev_trs_def)
    (*FIXME: default value for normal-form substitutions*)
  from string_reversal_qrstep_SN_rel[OF unary this]
  show "SN_qrel (qreltrs tp)"
    unfolding qreltrs_sound .
qed
end

end

